@@ -4,6 +4,7 @@ Generic pipelines for datasets
 
 import json
 import os
+import copy
 import random
 import shutil
 import tarfile
@@ -11,8 +12,11 @@ from datetime import datetime
 from pathlib import Path
 from typing import Dict, List, Optional, Set, Union
 from urllib.parse import urlparse
+import warnings
 
 import luigi
+import patoolib
+import numpy as np
 import pandas as pd
 from slugify import slugify
 from tqdm import tqdm
@@ -24,11 +28,13 @@ from hearpreprocess.util.luigi import (
     diagnostics,
     download_file,
     new_basedir,
+    safecopy,
     str2int,
 )
 
 INCLUDE_DATESTR_IN_FINAL_PATHS = False
 
+# Defaults for certain pipeline parameters
 SPLITS = ["train", "valid", "test"]
 # This percentage should not be changed as this decides
 # the data in the split and hence is not a part of the data config
@@ -38,22 +44,15 @@ TEST_PERCENTAGE = 10
 TRAIN_PERCENTAGE = 100 - VALIDATION_PERCENTAGE - TEST_PERCENTAGE
 TRAINVAL_PERCENTAGE = TRAIN_PERCENTAGE + VALIDATION_PERCENTAGE
 
-# We want no more than 5 hours of audio (training + validation) per task.
-# This can be overriden in the task config.
-# e.g. speech_commands test set.
-# If None, no limit is used.
-MAX_TASK_DURATION_BY_SPLIT = {
-    "train": 3600 * 5 * TRAIN_PERCENTAGE / TRAINVAL_PERCENTAGE,
-    "valid": 3600 * 5 * VALIDATION_PERCENTAGE / TRAINVAL_PERCENTAGE,
-    "test": 3600 * 5 * TEST_PERCENTAGE / TRAINVAL_PERCENTAGE,
-}
 
-
-def _diagnose_split_labels(taskname: str, event_str: str, df: pd.DataFrame):
+def _diagnose_split_labels(
+    task_config: Dict, taskname: str, event_str: str, df: pd.DataFrame
+):
     """Makes split and label diagnostics"""
     assert "split" in df.columns
     assert "label" in df.columns
 
+    splits = df["split"].unique()
     split_file_count = df.groupby("split")["relpath"].nunique().to_dict()
     # Get fraction of rows with a particular label for each split
     split_label_frac = {
@@ -70,35 +69,53 @@ def _diagnose_split_labels(taskname: str, event_str: str, df: pd.DataFrame):
         split: set(df["label"].unique()) - set(labels)
         for split, labels in df.groupby("split")["label"].apply(set).to_dict().items()
     }
-    for split in SPLITS:
-        if split in split_file_count:
-            diagnostics.info(
-                "{} {} file count {:5s}: {}".format(
-                    taskname, event_str, split, split_file_count[split]
-                )
+    for split in splits:
+        diagnostics.info(
+            "{} {} file count {:5s}: {}".format(
+                taskname, event_str, split, split_file_count[split]
             )
-    for split in SPLITS:
-        if split in split_label_freq:
-            diagnostics.info(
-                "{} {} label freq (descending) {:5s}: {}".format(
-                    taskname, event_str, split, split_label_freq[split]
-                )
+        )
+    for split in splits:
+        diagnostics.info(
+            "{} {} label freq (descending) {:5s}: {}".format(
+                taskname, event_str, split, split_label_freq[split]
             )
-    for split in SPLITS:
-        if split in split_label_freq:
-            diagnostics.info(
-                "{} {} label freq (alphabetical) {:5s}: {}".format(
-                    taskname, event_str, split, sorted(split_label_freq[split])
-                )
+        )
+    for split in splits:
+        diagnostics.info(
+            "{} {} label freq (alphabetical) {:5s}: {}".format(
+                taskname, event_str, split, sorted(split_label_freq[split])
             )
-    for split in SPLITS:
-        if split in split_label_missing and split_label_missing[split]:
+        )
+
+    for split in splits:
+        if split_label_missing[split]:
             diagnostics.info(
                 "{} {} MISSING LABELS {:5s}: {}".format(
                     taskname, event_str, split, split_label_missing[split]
                 )
             )
 
+    # Confirm that there are examples for all class labels
+    # This is a requirement for many metrics
+    if task_config["mode"] != "small" and any(
+        split_label_missing[split] for split in splits
+    ):
+        # Event multilabel tasks are an exception (e.g. maestro)
+        if (
+            task_config["embedding_type"] == "event"
+            and task_config["prediction_type"] == "multilabel"
+        ):
+            warnings.warn(
+                "All labels are not present across the splits. "
+                "Please check logs to see which files are missing."
+            )
+        else:
+            raise AssertionError(
+                "All labels are not present across the splits. "
+                "Please check logs to see which files are missing."
+            )
+
 
 class DownloadCorpus(WorkTask):
     """
@@ -151,8 +168,16 @@ class ExtractArchive(WorkTask):
     def run(self):
         archive_path = self.requires()["download"].workdir.joinpath(self.infile)
         archive_path = archive_path.absolute()
-        shutil.unpack_archive(archive_path, self.output_path)
-
+        # https://stackoverflow.com/questions/17614467/how-can-unrar-a-file-with-python
+        # Requires unrar for rar files - sudo apt install unrar
+        # Also works for types supported by shutil.unpack_archive
+        if self.output_path.exists():
+            shutil.rmtree(self.output_path)
+        self.output_path.mkdir(parents=True, exist_ok=True)
+        print(f"Extracting {self.infile} ..Please wait")
+        patoolib.extract_archive(
+            archive=str(archive_path), verbosity=0, outdir=str(self.output_path)
+        )
         self.mark_complete()
 
 
@@ -530,19 +555,113 @@ class ExtractMetadata(WorkTask):
         metadata.loc[metadata["split_key"].isin(test_split_keys), "split"] = "test"
         return metadata
 
+    def assert_correct_kfolds(self, metadata: pd.DataFrame):
+        """
+        Raises an AssertionError if the set of fold names in task_config doesn't
+        match the set of fold names used for splits in the metadata
+        """
+        if set(metadata["split"].unique()) != set(self.task_config["splits"]):
+            raise AssertionError(
+                "Names of splits in metadata don't match the required names for a "
+                f"kfold dataset. Expected: {self.task_config['splits']}. "
+                f"Received: {metadata['split'].unique()}."
+            )
+
+    def split_k_folds(self, metadata: pd.DataFrame):
+        """
+        Deterministically split dataset into k-folds
+        """
+        SEED_RETRIES = 1000
+        # Deterministically sort all unique split_keys.
+        split_keys = sorted(metadata["split_key"].unique())
+
+        # A set of seeds will be tried to ensure each fold has at least
+        # one example for each label
+        # Since seeds are tried sequentially, this is still deterministic
+        # The first seed remains "split_k_folds" to maintain consistency with
+        # previous versions
+        seeds = ["split_k_folds"] + [
+            f"split_k_folds_retry_{i}" for i in range(SEED_RETRIES - 1)
+        ]
+
+        for i, seed in enumerate(seeds):
+            # Deterministically shuffle all unique split_keys.
+            rng = random.Random(seed)
+            shuffled_split_keys = copy.deepcopy(split_keys)
+            rng.shuffle(shuffled_split_keys)
+
+            # Equally split the split_keys into k folds and label accordingly
+            k_folds = self.task_config["nfolds"]
+            folds_keys = np.array_split(shuffled_split_keys, k_folds)
+            for j, fold in enumerate(folds_keys):
+                metadata.loc[metadata["split_key"].isin(fold), "split"] = f"fold{j:02d}"
+
+            if self.task_config["mode"] == "small" or (
+                self.task_config["embedding_type"] == "event"
+                and self.task_config["prediction_type"] == "multilabel"
+            ):
+                # All labels are not required across all folds for either small
+                # or an event and multilable type task
+                break
+
+            if all(
+                metadata.groupby("split")["label"].nunique()
+                == metadata["label"].nunique()
+            ):
+                diagnostics.info(
+                    f"{self.longname} - Kfold was successful with seed: {seed}"
+                )
+                break
+
+        if i == len(seeds) - 1:
+            raise AssertionError(
+                "All the seeds were tried, but the data couldnot be split into "
+                "folds with at least one example for each label"
+            )
+        return metadata
+
+    def create_splits(self, metadata: pd.DataFrame) -> pd.DataFrame:
+        """
+        Splits the dataset based on the split mode in task config.
+        Either train/validation/test split or a k-fold split
+        """
+        if self.task_config["split_mode"] == "trainvaltest":
+            # Split the metadata to create valid and test set from train if
+            # they are not created explicitly in get_all_metadata
+            if set(self.task_config["splits"]) != set(SPLITS):
+                raise AssertionError(f"Splits for trainvaltest mode must be {SPLITS}")
+            metadata = self.split_train_test_val(metadata)
+        elif self.task_config["split_mode"] == "presplit_kfold":
+            # Splits are the predefined folds in the dataset - just make sure the
+            # names are correct
+            self.assert_correct_kfolds(metadata)
+        elif self.task_config["split_mode"] == "new_split_kfold":
+            # Split the dataset into k-folds
+            metadata = self.split_k_folds(metadata)
+            self.assert_correct_kfolds(metadata)
+        else:
+            raise ValueError("Unknown split_mode received in task_config")
+
+        return metadata
+
     def trim_event_metadata(self, metadata: pd.DataFrame, duration: float):
+        """
+        This modifies the event metadata to
+        retain only events which start before the sample duration
+        and trim them if they extend beyond the sample duration
+        """
         # Since the duration in the task config is in seconds convert to milliseconds
         duration_ms = duration * 1000.0
         assert "start" in metadata.columns
         assert "end" in metadata.columns
 
-        # Drop the events starting after the sample duration
-        trimmed_metadata = metadata.loc[lambda df: df["start"] < duration_ms]
+        # Drop the events starting at or after the sample duration
+        trimmed_metadata = metadata.loc[metadata["start"] < duration_ms]
         events_dropped = len(metadata) - len(trimmed_metadata)
 
         # Trim the events starting before but extending beyond the sample duration
-        events_trimmed = len(trimmed_metadata.loc[lambda df: df["end"] > duration_ms])
-        trimmed_metadata.loc[lambda df: df["end"] > duration_ms] = duration_ms
+        events_trimmed = len(trimmed_metadata.loc[metadata["end"] > duration_ms])
+        trimmed_metadata.loc[metadata["end"] > duration_ms, "end"] = duration_ms
 
         assert (trimmed_metadata["start"] < duration_ms).all()
         assert (trimmed_metadata["end"] <= duration_ms).all()
@@ -585,20 +704,22 @@ class ExtractMetadata(WorkTask):
         metadata = self.get_all_metadata()
         print(f"metadata length = {len(metadata)}")
 
-        _diagnose_split_labels(self.longname, "original", metadata)
+        _diagnose_split_labels(self.task_config, self.longname, "original", metadata)
         metadata = self.postprocess_all_metadata(metadata)
-        _diagnose_split_labels(self.longname, "postprocessed", metadata)
+        _diagnose_split_labels(
+            self.task_config, self.longname, "postprocessed", metadata
+        )
 
-        # Split the metadata to create valid and test set from train if they are not
-        # created explicitly in get_all_metadata
-        metadata = self.split_train_test_val(metadata)
+        # Creates splits based on the split_mode for this dataset
+        # Either train/val/test split or a k-fold split strategy
+        metadata = self.create_splits(metadata)
 
         # Each split should have unique files and no file should be across splits
         assert (
             metadata.groupby("unique_filestem")["split"].nunique() == 1
         ).all(), "One unique_filestem is associated with more than one split"
 
-        _diagnose_split_labels(self.longname, "split", metadata)
+        _diagnose_split_labels(self.task_config, self.longname, "split", metadata)
 
         if self.task_config["embedding_type"] == "scene":
             # Multiclass predictions should only have a single label per file
@@ -613,9 +734,17 @@ class ExtractMetadata(WorkTask):
             # duration
             # sample duration is specified in the task config.
             # The specified sample duration is in seconds
-            metadata = self.trim_event_metadata(
-                metadata, duration=self.task_config["sample_duration"]
-            )
+
+            # If the sample duration is set to None, no trimming of events will
+            # be done and the full audio file will be selected. This mode is
+            # only for special tasks and should not be generally used.
+            # Having all the audio files of the same length is more
+            # efficient for downstream pipelines
+
+            if self.task_config["sample_duration"] is not None:
+                metadata = self.trim_event_metadata(
+                    metadata, duration=self.task_config["sample_duration"]
+                )
         else:
             raise ValueError(
                 "%s embedding_type unknown" % self.task_config["embedding_type"]
@@ -676,7 +805,8 @@ class MetadataTask(WorkTask):
     def metadata(self):
         if self._metadata is None:
             self._metadata = pd.read_csv(
-                self.metadata_task.workdir.joinpath(self.metadata_task.outfile)
+                self.metadata_task.workdir.joinpath(self.metadata_task.outfile),
+                dtype={"relpath": str, "unique_filestem": str, "split": str},
             )
         return self._metadata
 
@@ -717,6 +847,28 @@ class SubsampleSplit(SplitTask):
             "metadata": self.metadata_task,
         }
 
+    def get_max_split_duration(self) -> Union[float, int, None]:
+        """
+        Returns the max duration for the current split from the task_config
+        """
+        # Key to use to get the max task duration depending on the split mode
+        if self.task_config["split_mode"] == "trainvaltest":
+            assert "max_task_duration_by_split" in self.task_config
+            max_durations = self.task_config["max_task_duration_by_split"]
+            if set(max_durations.keys()) != set(self.task_config["splits"]):
+                raise AssertionError(
+                    "Max duration must be specified for all splits/folds in "
+                    "task_config, or set to None to use the full length."
+                    f"Expected: {set(self.task_config['splits'])}, received:"
+                    f"{set(max_durations.keys())}."
+                )
+            max_split_duration = max_durations[self.split]
+        else:
+            assert "max_task_duration_by_fold" in self.task_config
+            max_split_duration = self.task_config["max_task_duration_by_fold"]
+
+        return max_split_duration
+
     def run(self):
         self.createsplit()
 
@@ -742,20 +894,24 @@ class SubsampleSplit(SplitTask):
         # But we aren't going to use audio that is more than a couple
         # minutes or the timestamp embeddings will explode
         sample_duration = self.task_config["sample_duration"]
+        max_split_duration = self.get_max_split_duration()
+        if sample_duration is None:
+            assert max_split_duration is None, (
+                "If the sample duration is set to None i.e. orignal audio files "
+                "are being used without any trimming or padding, then the "
+                "max_split_duration should also be None, so that no "
+                "subsampling is done as the audio file length is not "
+                "consistent."
+            )
 
-        if (
-            "max_task_duration_by_split" in self.task_config
-            and self.split in self.task_config["max_task_duration_by_split"]
-        ):
-            max_split_duration = self.task_config["max_task_duration_by_split"][
-                self.split
-            ]
-        else:
-            max_split_duration = MAX_TASK_DURATION_BY_SPLIT[self.split]
-        if max_split_duration is None:
-            max_files = len(split_filestem_relpaths)
-        else:
+        # If max_split_duration is not None set the max_files so that
+        # the total duration of all the audio files after subsampling
+        # comes around (less than) max_split_duration
+        if max_split_duration is not None:
             max_files = int(max_split_duration / sample_duration)
+        else:
+            # Otherwise set max_files to select all the files in the split
+            max_files = len(split_filestem_relpaths)
 
         diagnostics.info(
             f"{self.longname} "
@@ -848,15 +1004,18 @@ class TrimPadSplit(SplitTask):
 
     def run(self):
         self.createsplit()
-
         for audiofile in tqdm(list(self.requires()["corpus"].splitdir.iterdir())):
             newaudiofile = self.splitdir.joinpath(f"{audiofile.stem}.wav")
-            audio_util.trim_pad_wav(
-                str(audiofile),
-                str(newaudiofile),
-                duration=self.task_config["sample_duration"],
-            )
-
+            if self.task_config["sample_duration"] is not None:
+                audio_util.trim_pad_wav(
+                    str(audiofile),
+                    str(newaudiofile),
+                    duration=self.task_config["sample_duration"],
+                )
+            else:
+                # If the sample_duration is None, the file will be copied
+                # without any trimming or padding
+                safecopy(src=audiofile, dst=newaudiofile)
         self.mark_complete()
 
 
@@ -877,7 +1036,7 @@ class SubcorpusData(MetadataTask):
                 metadata_task=self.metadata_task,
                 task_config=self.task_config,
             )
-            for split in SPLITS
+            for split in self.task_config["splits"]
         }
         return splits
 
@@ -897,7 +1056,7 @@ class SubcorpusData(MetadataTask):
             assert self.requires()[key].workdir == self.requires()[key2].workdir
 
         # Output stats for every input directory
-        for split in SPLITS:
+        for split in self.task_config["splits"]:
             stats = audio_util.get_audio_dir_stats(
                 in_dir=self.workdir.joinpath(split),
                 out_file=self.workdir.joinpath(f"{split}_stats.json"),
@@ -925,7 +1084,7 @@ class SubcorpusMetadata(MetadataTask):
 
     def run(self):
         split_label_dfs = []
-        for split in SPLITS:
+        for split in self.task_config["splits"]:
             split_path = self.requires()["data"].workdir.joinpath(split)
             audiodf = pd.DataFrame(
                 [(a.stem, a.suffix) for a in list(split_path.glob("*.wav"))],
@@ -983,7 +1142,10 @@ class SubcorpusMetadata(MetadataTask):
             split_label_dfs.append(audiolabel_df)
 
         _diagnose_split_labels(
-            self.longname, "", pd.concat(split_label_dfs).reset_index(drop=True)
+            self.task_config,
+            self.longname,
+            "",
+            pd.concat(split_label_dfs).reset_index(drop=True),
         )
         self.mark_complete()
 
@@ -1007,7 +1169,7 @@ class MetadataVocabulary(MetadataTask):
     def run(self):
         labelset = set()
         # Save statistics about each subcorpus metadata
-        for split in SPLITS:
+        for split in self.task_config["splits"]:
             labeldf = pd.read_csv(
                 self.requires()["subcorpus_metadata"].workdir.joinpath(f"{split}.csv")
             )
@@ -1087,7 +1249,7 @@ class ResampleSubcorpuses(MetadataTask):
                 task_config=self.task_config,
             )
             for sr in self.sample_rates
-            for split in SPLITS
+            for split in self.task_config["splits"]
         ]
         return resample_splits
 
@@ -1237,7 +1399,7 @@ class TarCorpus(MetadataTask):
 
         # Compute the audio files to be tar'ed
         files = set()
-        for split in SPLITS:
+        for split in self.task_config["splits"]:
             files |= set(
                 json.load(open(os.path.join(source_dir, f"{split}.json"))).keys()
             )
